<?php

namespace Drupal\config_auto_export;

use Drupal\Component\Datetime\Time;
use Drupal\config\StorageReplaceDataWrapper;
use Drupal\Core\Config\CachedStorage;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Config\ConfigImporter;
use Drupal\Core\Config\ConfigManager;
use Drupal\Core\Config\StorageComparer;
use Drupal\Core\Config\TypedConfigManager;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Extension\ModuleExtensionList;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Extension\ModuleInstallerInterface;
use Drupal\Core\Extension\ThemeHandler;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Http\ClientFactory;
use Drupal\Core\Lock\LockBackendInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\Serialization\Yaml;
use Drupal\Core\Site\Settings;
use Drupal\Core\State\StateInterface;
use Drupal\Core\StringTranslation\TranslationManager;
use GuzzleHttp\Exception\GuzzleException;
use Drupal\Core\Session\AccountInterface;
use Symfony\Contracts\EventDispatcher\EventDispatcherInterface;

class Service {

  public const STATE_KEY_DUE_TIMESTAMP = 'config_auto_export.due_next.timestamp';

  public const STATE_KEY_PAUSED = 'config_auto_export.paused';

  /**
   * @var \Drupal\Core\Config\ImmutableConfig
   */
  protected $config;

  /**
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * @var \Drupal\Core\File\FileSystemInterface
   */
  protected $fileSystem;

  /**
   * @var \Drupal\Core\State\StateInterface
   */
  protected $state;

  /**
   * @var \Drupal\Component\Datetime\Time
   */
  protected $time;

  /**
   * @var \Drupal\Core\Http\ClientFactory
   */
  protected $client;

  /**
   * @var \Drupal\Core\Logger\LoggerChannelInterface
   */
  protected $logger;

  /**
   * @var \Drupal\Core\Session\AccountInterface
   */
  protected $account;

  /**
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected $moduleHandler;

  /**
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * @var \Drupal\Core\Config\CachedStorage
   */
  protected $storage;

  /**
   * @var \Symfony\Contracts\EventDispatcher\EventDispatcherInterface
   */
  protected $eventDispatcher;

  /**
   * @var \Drupal\Core\Config\ConfigManager
   */
  protected $configManager;

  /**
   * @var \Drupal\Core\Lock\LockBackendInterface
   */
  protected $lock;

  /**
   * @var \Drupal\Core\Config\TypedConfigManager
   */
  protected $configTyped;

  /**
   * @var \Drupal\Core\Extension\ModuleInstallerInterface
   */
  protected $moduleInstaller;

  /**
   * @var \Drupal\Core\Extension\ThemeHandler
   */
  protected $themeHandler;

  /**
   * @var \Drupal\Core\StringTranslation\TranslationManager
   */
  protected $stringTranslation;

  /**
   * @var \Drupal\Core\Extension\ModuleExtensionList
   */
  protected $extensionListModule;

  /**
   * Service constructor.
   *
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   * @param \Drupal\Core\File\FileSystemInterface $file_system
   * @param \Drupal\Core\Logger\LoggerChannelFactoryInterface $logger_channel_factory
   * @param \Drupal\Core\State\StateInterface $state
   * @param \Drupal\Component\Datetime\Time $time
   * @param \Drupal\Core\Http\ClientFactory $client
   * @param \Drupal\Core\Session\AccountInterface $account
   * @param \Drupal\Core\Extension\ModuleHandlerInterface $module_handler
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   * @param \Drupal\Core\Config\CachedStorage $storage
   * @param \Drupal\Component\EventDispatcher\ContainerAwareEventDispatcher $event_dispatcher
   * @param \Drupal\Core\Config\ConfigManager $config_manager
   * @param \Drupal\Core\Lock\LockBackendInterface $lock
   * @param \Drupal\Core\Config\TypedConfigManager $config_typed
   * @param \Drupal\Core\Extension\ModuleInstallerInterface $module_installer
   * @param \Drupal\Core\Extension\ThemeHandler $theme_handler
   * @param \Drupal\Core\StringTranslation\TranslationManager $string_translation
   * @param \Drupal\Core\Extension\ModuleExtensionList $extension_list_module
   */
  public function __construct(ConfigFactoryInterface $config_factory, FileSystemInterface $file_system,
                              LoggerChannelFactoryInterface $logger_channel_factory, StateInterface $state,
                              Time $time, ClientFactory $client, AccountInterface $account,
                              ModuleHandlerInterface $module_handler, EntityTypeManagerInterface $entity_type_manager,
                              CachedStorage $storage, EventDispatcherInterface $event_dispatcher, ConfigManager $config_manager,
                              LockBackendInterface $lock, TypedConfigManager $config_typed, ModuleInstallerInterface $module_installer,
                              ThemeHandler $theme_handler, TranslationManager $string_translation, ModuleExtensionList $extension_list_module) {
    $this->config = $config_factory->get('config_auto_export.settings');
    $this->configFactory = $config_factory;
    $this->fileSystem = $file_system;
    $this->logger = $logger_channel_factory->get('config');
    $this->state = $state;
    $this->time = $time;
    $this->client = $client;
    $this->account = $account;
    $this->moduleHandler = $module_handler;
    $this->entityTypeManager = $entity_type_manager;
    $this->storage = $storage;
    $this->eventDispatcher = $event_dispatcher;
    $this->configManager = $config_manager;
    $this->lock = $lock;
    $this->configTyped = $config_typed;
    $this->moduleInstaller = $module_installer;
    $this->themeHandler = $theme_handler;
    $this->stringTranslation = $string_translation;
    $this->extensionListModule = $extension_list_module;
  }

  /**
   * Pause config auto export.
   */
  public function pause(): void {
    $this->state->set(self::STATE_KEY_PAUSED, TRUE);
  }

  /**
   * Resume config auto export.
   */
  public function resume(): void {
    $this->state->set(self::STATE_KEY_PAUSED, FALSE);
  }

  /**
   * Is the process currently paused?
   *
   * @return bool
   */
  public function isPaused(): bool {
    return $this->state->get(self::STATE_KEY_PAUSED, FALSE);
  }

  /**
   * Trigger the webhook.
   *
   * @param bool $force
   *   Force trigger even if service is paused.
   *
   * @return bool
   */
  public function triggerExport(bool $force = FALSE): bool {
    if (!$force && $this->isPaused()) {
      return FALSE;
    }
    if (!$this->config->get('enabled')) {
      return FALSE;
    }
    $webhook = $this->config->get('webhook');
    if (empty($webhook)) {
      return FALSE;
    }
    $exportPath = $this->fileSystem->realpath($this->config->get('directory'));
    $configPath = $this->fileSystem->realpath(Settings::get('config_sync_directory'));
    $currentUser = $this->account->getDisplayName();
    $configSplitPaths = [];
    if ($this->moduleHandler->moduleExists('config_split')) {
      /** @var \Drupal\config_split\Entity\ConfigSplitEntityInterface $split */
      foreach ($this->entityTypeManager->getStorage('config_split')
                 ->loadMultiple() as $split) {
        if ($split->status()) {
          $configSplitPaths[$split->id()] = $split->get('folder');
        }
      }
    }
    $data = [
      'form_params' => Yaml::decode(str_replace(
        [
          '[current user]',
          '[export directory]',
          '[config directory]',
          '[config split directories]',
        ],
        [
          $currentUser,
          $exportPath,
          $configPath,
          json_encode($configSplitPaths),
        ],
        $this->config->get('webhook_params'))),
    ];

    try {
      $client = $this->client->fromOptions(['base_uri' => $webhook]);
      $client->request('post', '', $data);
    } catch (GuzzleException $e) {
      $this->logger->critical('Trigger for config auto export failed: {msg}', ['msg' => $e->getMessage()]);
      return FALSE;
    }
    $this->state->delete(self::STATE_KEY_DUE_TIMESTAMP);
    return TRUE;
  }

  /**
   * Check if a trigger due date exists and is due, then call the trigger.
   */
  public function checkDueDate(): void {
    if (($dueTime = $this->state->get(self::STATE_KEY_DUE_TIMESTAMP)) && $dueTime <= $this->time->getRequestTime()) {
      $this->triggerExport();
    }
  }

  /**
   * Reload a config file from a module's install or optional directory.
   *
   * @param string $module
   * @param string $section
   * @param string[] $configIds
   */
  public function reloadConfig(string $module, string $section, array $configIds): void {
    if (!$this->moduleHandler->moduleExists($module)) {
      $this->logger->alert('Reload config was called for %module, which is not enabled.', [
        '%module' => $module,
      ]);
      return;
    }
    if (!in_array($section, ['install', 'optional', 'overwrite'])) {
      $this->logger->alert('Reload config was called for %module with section %section, which is not allowed.', [
        '%module' => $module,
        '%section' => $section,
      ]);
      return;
    }

    foreach ($configIds as $configId) {
      $filename = drupal_get_path('module', $module) . "/config/$section/$configId.yml";
      if (!file_exists($filename)) {
        $this->logger->alert('Reload config was called for %module with section %section and config ID %configid, which does not exist.', [
          '%module' => $module,
          '%section' => $section,
          '%configid' => $configId,
        ]);
        continue;
      }
      $data = Yaml::decode(file_get_contents($filename));
      if ($config = $this->configFactory->getEditable($configId)) {
        if ($uuid = $config->get('uuid')) {
          $data['uuid'] = $uuid;
        }
        $config->setData($data)->save();
      }
      else {
        $this->createConfig($configId, $data);
      }
    }
  }

  /**
   * @param string $configId
   * @param array $data
   */
  public function createConfig(string $configId, array $data): void {
    $sourceStorage = new StorageReplaceDataWrapper($this->storage);
    $sourceStorage->delete($configId);
    $sourceStorage->write($configId, $data);
    $storageComparer = new StorageComparer($sourceStorage, $this->storage);
    $configImporter = new ConfigImporter(
      $storageComparer,
      $this->eventDispatcher,
      $this->configManager,
      $this->lock,
      $this->configTyped,
      $this->moduleHandler,
      $this->moduleInstaller,
      $this->themeHandler,
      $this->stringTranslation,
      $this->extensionListModule,
    );
    if ($configImporter->alreadyImporting()) {
      $this->logger->alert('Import already running.');
      return;
    }
    if (!$configImporter->validate()) {
      $this->logger->alert('To be imported config not valid.');
      return;
    }
    $sync_steps = $configImporter->initialize();
    foreach ($sync_steps as $step) {
      $context = [];
      do {
        $configImporter->doSyncStep($step, $context);
      } while ($context['finished'] < 1);
    }
  }

}
